local CM = ChatMonitor
local math, string = math, string
local type, min, table = type, min, table
local ipairs = ipairs
local L = CM.L

local function getWords(msg)
	local words = {}
	local word
	for word in msg:gmatch("%w+") do table.insert(words, word) end
	return words
end

local function unlink(msg)
	--replace links with the link text
	local ul = string.gsub(msg, "|c.-%[", "")
	ul = string.gsub(ul, "%]|h|r", "")
	return ul
end

--*************************************************
--* DamerauLevenshtein Distance                  *
--* based on code from http://nayruden.com/?p=115 *
--* s - string/table                              *
--* t - string/table to cpmare against            *
--* lim - limit to a maximum distance             *
--*************************************************
local function getLevenshtein(s, t, lim)
    local slen, tlen = #s, #t
    if lim and math.abs(slen - tlen) >= lim then return lim end
    if type(s) == "string" then s = {string.byte(s, 1, slen)} end
    if type(t) == "string" then t = {string.byte(t, 1, tlen)} end
    local numcolumns = tlen + 1
    local d = {}
    for i = 0, slen do d[i * numcolumns] = i end
    for j = 0, tlen do d[j] = j end
    for i = 1, slen do
        local ipos = i * numcolumns
        local best = lim
        for j = 1, tlen do
            local addcost = (s[i] ~= t[j] and 1 or 0)
            local val = min(
                d[ipos - numcolumns + j] + 1,
                d[ipos + j - 1] + 1,
                d[ipos - numcolumns + j - 1] + addcost)
            d[ipos + j] = val
            if i > 1 and j > 1 and s[i] == t[j - 1] and s[i - 1] == t[j] then
                d[ipos + j] = min(val, d[ipos - numcolumns - numcolumns + j - 2] + addcost)
            end
            if lim and val < best then best = val end
        end
        if lim and best >= lim then return lim end
    end
    return d[#d]
end

function CM:FilterChat(msg, author, language, channamenum, target, flags, zoneid, channum, channame, lineid, senderguid)
	if not CM.alert:IsVisible() then return end
	local terms, distance, found
	local limit = CM.db.profile.maxdistance
	if  (CM.db.char.changeneral and zoneid == 1) or 
		(CM.db.char.chantrade and zoneid == 2) or
		(CM.db.char.chanlfg and zoneid == 4) or
		(CM.db.char.changuild and zoneid == 0) then
		local clearmsg = string.lower(unlink(msg))
		local words = getWords(clearmsg)
		local filters = {"wtb","lfg","wts","js","enc","lw","ins","tlr","bs","alc","eng",
			"deathknight","warrior","rogue","mage","priest","hunter","warlock","demonhunter",
			"shaman","druid","monk","paladin","heal","dps","mel","rng","cc","tank","guild",
			"lfgc","cus","lfguild","lfm"}
		terms = {}

		for _, filter in ipairs(filters) do
			if CM.db.char[filter] then
				table.insert(terms, {
					keywords = {strsplit(",", CM.db.profile[format("%slist", filter)])}, 
					sound = CM.db.profile[format("%ssound", filter)],
					colour = CM.db.profile[format("%scolour", filter)],
					type = filter})
			end
		end

		for _, word in ipairs(words) do
			for _, termgroup in ipairs(terms) do
				for _, term in ipairs(termgroup.keywords) do
					local t = string.lower(term)
					local dlim = limit
					local isreg = (string.sub(t, 1, 1) == "/") and (string.sub(t, -1) == "/")
					
					if isreg then
						t = string.sub(t, 2)
						t = string.sub(t, 1, string.len(t) - 1)
						if termgroup.type == "lfguild" then
							if string.find(clearmsg, "<.+>") or string.find(clearmsg, ">.+<") or
								string.find(clearmsg, "/w") then
								-- this is more likely an ad rather than someone looking for a guild
								found = true
								break
							end
						end
						if not found then
							if string.find(clearmsg, t) then
								distance = true
							else
								distance = false
							end
							dlim = false
						end
					elseif string.len(t) < 4 then
						distance = word == t
						dlim = false
					else
						distance = getLevenshtein(word, string.lower(term), limit)
					end
					if distance ~= dlim then
						CM:AddAlert(msg, author, senderguid, termgroup.type, termgroup.sound, termgroup.colour, zoneid)
						found = true
						break
					end
				end
				if found then break end
			end
			if found then break end
		end
	end
end